"""Translate ODEs in the YAML format into PEtab."""
import argparse
import os
import warnings

import libsbml as sbml
import pandas as pd
import petab
import yaml
from pathlib import Path


from .yaml2sbml import _parse_yaml_dict, _load_yaml_file
from .yaml_validation import _validate_yaml_from_dict


def yaml2petab(yaml_dir: str,
               output_dir: str,
               sbml_name: str,
               petab_yaml_name: str = None,
               measurement_table_name: str = None):
    """
    Translate a YAML model into a PEtab model.

    Takes a YAML file with the ODE specification, parses it, converts
    it into SBML format, and writes the SBML file. Further it translates the
    given information into PEtab tables.

    If a `petab_yaml_name` is given, a YAML file is created, that organizes
    the PEtab problem. If additionally a `measurement_table_file_name` is
    specified, this file name is written into the created YAML file.

    Arguments:
        yaml_dir : path to the YAML file with the ODEs specification
        output_dir: path the output file(s) are be written out
        sbml_name: name of SBML model
        petab_yaml_name: name of YAML organizing the PEtab problem.
        measurement_table_name: Name of measurement table
    """
    yaml_model_dict = _load_yaml_file(yaml_dir)
    _yaml2petab(yaml_model_dict,
                output_dir,
                sbml_name,
                petab_yaml_name,
                measurement_table_name)


def _yaml2petab(yaml_model_dict: dict,
                output_dir: str,
                model_name: str,
                petab_yaml_name: str = None,
                measurement_table_name: str = None):
    """
    Similar to 'yaml2petab', but takes a yaml_model_dict as input.

    If a petab_yaml_name is given, a .yaml file is created, that organizes
    the petab problem. If additionally a measurement_table_file_name is
    specified, this file name is written into the created .yaml file.

    Arguments:
        yaml_model_dict : dictionary, containing the yaml model
        output_dir: path the output file(s) are be written out
        model_name: name of SBML model
        petab_yaml_name: name of yaml organizing the PEtab problem.
        measurement_table_name: Name of measurement table
    """
    # validate yaml
    _validate_yaml_from_dict(yaml_model_dict)

    # output make directory, if it doesn't exist yet.
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    if model_name.endswith('.xml') or model_name.endswith('.sbml'):
        sbml_dir = os.path.join(output_dir, model_name)
    else:
        sbml_dir = os.path.join(output_dir, model_name + '.xml')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        sbml_as_string = _parse_yaml_dict(yaml_model_dict,
                                          model_name)

    with open(sbml_dir, 'w') as f_out:
        f_out.write(sbml_as_string)

    # create petab tsv files:
    if model_name.endswith('.xml') or model_name.endswith('.sbml'):
        model_name = Path(model_name).stem

    _create_petab_tables_from_yaml(yaml_model_dict,
                                   output_dir,
                                   model_name)

    # create yaml file, that organizes the petab problem:
    if (petab_yaml_name is None) and (measurement_table_name is not None):

        warnings.warn('Since no petab_yaml_file_name is specified, the '
                      'specified measurement_table_name will have no effect.',
                      RuntimeWarning)

    elif petab_yaml_name is not None:
        _create_petab_problem_yaml(yaml_model_dict,
                                   output_dir,
                                   sbml_dir,
                                   petab_yaml_name,
                                   model_name,
                                   measurement_table_name)
    # validate PEtab tables:
    validate_petab_tables(sbml_dir, output_dir)


def _create_petab_tables_from_yaml(yaml_dict: dict,
                                   output_dir: str,
                                   model_name: str):
    """
    Parse the YAML dict to a PEtab observable/parameter table.

    The files will be named (if the corresponding information is in the YAML):
    `parameters_<model_name>.tsv`,
    `observables_<model_name>.tsv` and
    `experimental_conditions_<model_name>.tsv`.

    Arguments:
        yaml_dict: dict, that contains the yaml file.
        output_dir: directory, where the PEtab tables should be written.
    """
    parameter_table = _create_parameter_table(yaml_dict)
    parameter_table.to_csv(os.path.join(output_dir,
                                        f'parameters_{model_name}.tsv'),
                           sep='\t',
                           index=False)

    # create PEtab observable table, if observables occur in the yaml file.
    if 'observables' in yaml_dict.keys():
        observable_table = _create_observable_table(yaml_dict)
        observable_table.to_csv(os.path.join(output_dir,
                                             f'observables_{model_name}.tsv'),
                                sep='\t',
                                index=False)

    # create PEtab condition table, if conditions occur in the yaml file.
    if 'conditions' in yaml_dict.keys():
        condition_table = _create_condition_table(yaml_dict)
        condition_table.to_csv(
            os.path.join(output_dir,
                         f'experimental_conditions_{model_name}.tsv'),
            sep='\t',
            index=False)


def _create_petab_problem_yaml(yaml_dict: dict,
                               output_dir: str,
                               sbml_dir: str,
                               petab_yaml_name: str,
                               model_name: str,
                               measurement_table_name: str = None):
    """
    Create a YAML file, that can be used for defining a PEtab problem.

    Arguments:
        yaml_dict: dict, that contains the YAML file.
        output_dir: directory, where the PEtab tables should be written.
        sbml_dir: directory of the SBML model.
        petab_yaml_name: name of file, where PEtab YAML is written.
        model_name: name of the model, in order to name the PEtab tables.
        measurement_table_name: directory of the  measurement table.
    """
    petab_yaml_dict = {'format_version': 1,
                       'parameter_file': f'parameters_{model_name}.tsv',
                       'problems': [{'sbml_files':
                                         [os.path.basename(sbml_dir)]}]}

    # fill the corresponding entries, if they are contained in the yaml/input.

    if 'observables' in yaml_dict.keys():
        petab_yaml_dict['problems'][0]['observable_files'] = \
            [f'observables_{model_name}.tsv']

    if 'conditions' in yaml_dict.keys():
        petab_yaml_dict['problems'][0]['condition_files'] = \
            [f'experimental_conditions_{model_name}.tsv']

    if measurement_table_name is not None:
        petab_yaml_dict['problems'][0]['measurement_files'] = \
            [measurement_table_name]

    petab_yaml_dir = os.path.join(output_dir, petab_yaml_name)

    with open(petab_yaml_dir, 'w') as file:
        yaml.dump(petab_yaml_dict, file)


def _create_parameter_table(yaml_dict: dict):
    """
    Create a parameter table from the parameter block in `yaml_dict.

    Arguments:`
        yaml_dict

    Returns:
        parameter_table: pandas data frame containing the parameter table.
    """
    return _create_petab_table(yaml_dict['parameters'],
                               petab.PARAMETER_DF_REQUIRED_COLS,
                               petab.PARAMETER_DF_OPTIONAL_COLS)


def _create_observable_table(yaml_dict: dict):
    """
    Create an observable table from the observable block in `yaml_dict`.

    Arguments:
        yaml_dict

    Returns:
        observable_table: pandas data frame containing the observable table.
            (if observable block is not empty, else None)
    """
    return _create_petab_table(yaml_dict['observables'],
                               petab.OBSERVABLE_DF_REQUIRED_COLS,
                               petab.OBSERVABLE_DF_OPTIONAL_COLS)


def _create_condition_table(yaml_dict: dict):
    """
    Create a condition table from the condition block in `yaml_dict.

    Arguments:
        yaml_dict

    Returns:
        condition_table: pandas data frame containing the condition table.
            (if condition block is not empty, else None)
    """
    mandatory_id_list = [petab.CONDITION_ID]

    optional_id_list = [petab.CONDITION_NAME] + \
                       [p['parameterId'] for p in yaml_dict['parameters']] + \
                       [ode['stateId'] for ode in yaml_dict['odes']]

    return _create_petab_table(yaml_dict['conditions'],
                               mandatory_id_list,
                               optional_id_list)


def validate_petab_tables(sbml_dir: str,
                          output_dir: str):
    """
    Validate the PEtab tables via `petab.lint`.

    Throws an error if the PEtab tables do not follow the PEtab format
    standard.

    Arguments:
        sbml_dir: directory of the sbml
        output_dir: output directory for petab files

    Raises:
        Errors are raised by lint, if PEtab files are invalid...
    """
    model = sbml.readSBML(sbml_dir).getModel()
    model_name = model.getId()

    parameter_file_dir = \
        os.path.join(output_dir, f'parameters_{model_name}.tsv')
    observable_file_dir = \
        os.path.join(output_dir, f'observables_{model_name}.tsv')
    condition_table_dir = \
        os.path.join(output_dir, f'experimental_conditions_{model_name}.tsv')

    # check observable table, if the table exists
    if os.path.exists(observable_file_dir):
        observable_df = pd.read_csv(observable_file_dir,
                                    sep='\t',
                                    index_col='observableId')
        petab.lint.check_observable_df(observable_df)

    else:
        observable_df = None

    # check condition table, if the table exists
    if os.path.exists(condition_table_dir):
        condition_df = pd.read_csv(condition_table_dir,
                                   sep='\t',
                                   index_col='conditionId')
        petab.lint.check_condition_df(condition_df, model)

    # check parameter table
    parameter_df = pd.read_csv(parameter_file_dir,
                               sep='\t',
                               index_col='parameterId')

    petab.lint.check_parameter_df(parameter_df,
                                  sbml_model=model,
                                  observable_df=observable_df)


def _create_petab_table(block_list: list,
                        mandatory_id_list: list,
                        optional_id_list: id):
    """
    Create a PEtab table from the block_list in the yaml_dict.

    Arguments:
        block_list: entry from yaml_dict.
        mandatory_id_list: list of mandatory ids in the PEtab table
        optional_id_list: list of optional ids in the PEtab table

    Returns:
        petab_table: pandas data frame containing the petab table.
    """
    petab_table = pd.DataFrame({col_id: [] for col_id in mandatory_id_list})

    for row in block_list:
        _petab_table_add_row(petab_table, row)

    # check if every column is part of PEtab standard.
    for col_name in petab_table.head():
        if not (col_name in mandatory_id_list or col_name in optional_id_list):
            warnings.warn(f'PEtab warning: {col_name} is not part of the '
                          f'PEtab standard and hence might have noe effect.')
    return petab_table


def _petab_table_add_row(petab_table: pd.DataFrame, row_dict: dict):
    """
    Add a row defined by row_dict to the petab_table given by petab_table.

    After the function call the last row of petab_table has the entries

        petab_table[-1, key] = row_dict[key]

    for all keys of row_dict.

    Arguments:
        petab_table: pd.DataFrame containing the current petab table.
        row_dict: dict, that contains the values of the new row.
    """
    n_rows = petab_table.shape[0]

    for key in row_dict.keys():
        petab_table.loc[n_rows + 1, key] = row_dict[key]


def main():
    """Command-Line Interface."""
    parser = argparse.ArgumentParser(description='Takes in an ODE model in '
                                     '.yaml and converts it to a PEtab file.')

    parser.add_argument('yaml_file', type=str, help='Input yaml directory')
    parser.add_argument('output_dir', type=str, help='Output directory')
    parser.add_argument('model_name', type=str,
                        help='name of created SBML model.')
    parser.add_argument('-y', '--petab_yaml', type=str,
                        help='Optional argument, creates a petab .yml')
    parser.add_argument('-m', '--measurement_table', type=str,
                        help='Optional argument, path to measurement table')

    args = parser.parse_args()

    print(f'Path to yaml file: {args.yaml_file}')
    print(f'Output directory: {args.output_dir}')
    print(f'Path to sbml/petab files: {args.model_name}')

    print('Converting...')

    yaml2petab(args.yaml_file,
               args.output_dir,
               args.model_name,
               args.petab_yaml,
               args.measurement_table)


if __name__ == '__main__':
    main()



"""Translate ODEs in the YAML format into SBML."""
import argparse
import warnings
from pathlib import Path

import libsbml as sbml
import yaml
from yaml.scanner import ScannerError

from .yaml_validation import _validate_yaml_from_dict


def yaml2sbml(yaml_dir: str,
              sbml_dir: str,
              observables_as_assignments: bool = False):
    """
    Parse a YAML file with the specification of ODEs and write it to SBML.

    SBML is written within this function. If `observables_as_assignments=True`,
    observables will be translated into parameter assignments of the form
    `observable_<observable_id>`.

    Arguments:
        yaml_dir: directory to the YAML file with the ODEs specification
        sbml_dir: directory to the SBML file to be written out
        observables_as_assignments: indicates whether there should be
            parameter assignments of the form `observable_<observable_id>`.
    """
    # check file extension in sbml_dir
    if not (sbml_dir.endswith('.xml') or sbml_dir.endswith('.sbml')):
        raise ValueError('sbml_dir should end with .xml or .sbml.')

    model_name = Path(sbml_dir).stem

    sbml_as_string = _parse_yaml(yaml_dir,
                                 model_name,
                                 observables_as_assignments)

    # write sbml file
    with open(sbml_dir, 'w') as f_out:
        f_out.write(sbml_as_string)


def _parse_yaml(yaml_dir: str,
                model_name: str,
                observables_as_assignments: bool = False) -> str:
    """
    Parse a YAML file with the specification of ODEs to SBML.

    The SBML is returned as string.

    Arguments:
        yaml_dir: path to the YAML file with the ODEs specification
        model_name: model name as specified in the SBML
        observables_as_assignments: indicates if observables should be
            translated into parameter assignments

    Returns:
        sbml_string: a string containing the ODEs in SBML format

    Raises:
        SystemExit
    """
    yaml_dict = _load_yaml_file(yaml_dir)
    _validate_yaml_from_dict(yaml_dict)

    sbml_string = _parse_yaml_dict(yaml_dict,
                                   model_name,
                                   observables_as_assignments)

    return sbml_string


def _parse_yaml_dict(yaml_dict: dict,
                     model_name: str,
                     observables_as_assignments: bool = False) -> str:
    """
    Generate a string, containing the SBML from a `yaml_dict.

    Arguments:
        yaml_dict: dictionary, containing to the YAML file with the ODEs
                   specification.`
        model_name: model name as specified in the SBML
        observables_as_assignments: indicates if observables should be
            translated into parameter assignments

    Returns:
        sbml_string: a string containing the ODEs in SBML format.

    """
    try:
        document = sbml.SBMLDocument(3, 1)
    except ValueError:
        raise SystemExit('Could not create SBMLDocument object')

    model = document.createModel()

    # remove file extension
    if model_name.endswith('.xml') or model_name.endswith('.sbml'):
        model_name = Path(model_name).stem

    model.setId(model_name)
    model.setName(model_name)

    model = _create_compartment(model)

    _convert_yaml_blocks_to_sbml(model,
                                 yaml_dict,
                                 observables_as_assignments)

    # check consistency and give warnings for errors in SBML:
    if document.checkConsistency():

        for error_num in range(document.getErrorLog().getNumErrors()):
            if not document.getErrorLog().getError(error_num).isWarning():
                warnings.warn(
                    document.getErrorLog().getError(error_num).getMessage(),
                    RuntimeWarning)

    sbml_string = sbml.writeSBMLToString(document)

    return sbml_string


def _create_compartment(model: sbml.Model):
    """
    Create a default compartment for the model.

    yaml2sbml doesn't support multiple compartments.

    Arguments:
        model: SBML model

    Returns:
        model: SBML model with added compartment
    """
    c = model.createCompartment()
    c.setId('Compartment')
    c.setConstant(True)
    c.setSize(1)

    return model


def _load_yaml_file(yaml_file: str) -> dict:
    """
    Load YAML file from a dictionary.

    Arguments:
        yaml_file: directory to the YAML model

    Returns:
        yaml_dic: dictionary with parsed YAML file contents
    Raises:
        RuntimeError, if YAML can not be parsed, e.g. due to incorrectly
            formatted entries
    """
    try:

        with open(yaml_file, 'r') as f_in:
            yaml_contents = f_in.read()
            yaml_dict = yaml.full_load(yaml_contents)

    except ScannerError:
        raise RuntimeError('YAML file can not be parsed due to a Scanner '
                           'Error. This commonly happens if formulas begin '
                           'with a minus. Please set them inside of brackets '
                           '"(...)" or quotation marks.')
    return yaml_dict


def _convert_yaml_blocks_to_sbml(model: sbml.Model,
                                 yaml_dict: dict,
                                 observables_as_assignments):
    """
    Convert each block in the YAML dictionary to SBML.

    Arguments:
        model: SBML model
        yaml_dict: dictionary with YAML contents

    Returns:
        model: SBML model with added entities
    """

    def _read_observables_with_assignments(model, block):
        return _read_observables_block(model,
                                       block,
                                       observables_as_assignments)

    function_dict = {'time': _read_time_block,
                     'parameters': _read_parameters_block,
                     'assignments': _read_assignments_block,
                     'functions': _read_functions_block,
                     'observables': _read_observables_with_assignments,
                     'odes': _read_odes_block,
                     'conditions': _read_conditions_block}

    for block in yaml_dict:
        function_dict[block](model, yaml_dict[block])

    return model


def _read_time_block(model: sbml.Model, time_dic: dict):
    """
    Read and process the time block.

    Arguments:
        model: SBML model to which the rate rule will be added.
        time_dic: a dictionary with the time block in the ODE YAML file.
    """
    if time_dic['variable'] == 'time':
        return
    else:
        _create_time(model, time_dic['variable'])


def _create_time(model: sbml.Model, time_var: str):
    """
    Create the time variable, add assignment to 'time'.

    Arguments:
        model: the SBML model to which the species will be added.
        time_var: str, the time variable
    """
    if time_var == 'time':
        return

    time_parameter = model.createParameter()

    if time_parameter.setId(time_var) != sbml.LIBSBML_OPERATION_SUCCESS:

        raise RuntimeError(f'Unable to generate time parameter with id '
                           f'{time_var}. Invalid SBML identifier.')

    time_parameter.setName(time_var)
    time_parameter.setConstant(False)

    time_assignment = model.createAssignmentRule()
    time_assignment.setVariable(time_var)
    time_assignment.setMath(sbml.parseL3Formula('time'))


def _read_parameters_block(model: sbml.Model, parameter_list: list):
    """
    Read and process the parameters block in the YAML file.

    The expected format for parameter definition is
    {'parameterId': <parameterId>, 'nominalValue': <nominalValue>}

    Arguments:
        model: the SBML model
        parameter_list: block containing the parameter definitions
    """
    for parameter_def in parameter_list:
        if 'nominalValue' in parameter_def.keys():
            _create_parameter(model,
                              parameter_def['parameterId'],
                              parameter_def['nominalValue'])
        else:
            _create_parameter(model, parameter_def['parameterId'])


def _create_parameter(model: sbml.Model, parameter_id: str, value: str = None):
    """
    Create a parameter and add it to the given SBML model.

    Units are set as dimensionless by default.

    Arguments:
        model: the SBML model to which the parameter will be added.
        parameter_id: the parameter ID
        value: the parameter value, if value is None, no parameter is set.
    """
    k = model.createParameter()

    if k.setId(parameter_id) != sbml.LIBSBML_OPERATION_SUCCESS:
        raise RuntimeError(f'Unable to generate parameter with id '
                           f'{parameter_id}. Invalid SBML identifier.')

    k.setName(parameter_id)
    k.setConstant(True)

    if value is not None:
        k.setValue(float(value))

    k.setUnits('dimensionless')


def _read_assignments_block(model: sbml.Model, assignment_list: list):
    """
    Read and process the assignments block in the YAML file.

    The expected format of a state definition is:
    {'assignmentId': <assignmentId>, 'formula': <formula>}

    This is used to assign a formula (possibly time-dependent) to a variable.

    Arguments:
        model: the SBML model
        assignment_list: a list of dictionaries where each entry is an
                         assignment definition
    """
    for assignment_def in assignment_list:
        _create_assignment(model,
                           assignment_def['assignmentId'],
                           assignment_def['formula'])


def _create_assignment(model: sbml.Model, assignment_id: str, formula: str):
    """
    Create an assignment rule, that assigns <id> to <formula>.

    Arguments:
        model: SBML model to which the assignment rule will be added.
        assignment_id: str, the id of the assignment rule
        formula: str: contains the equation for the assignment rule
    """
    assignment_parameter = model.createParameter()

    if assignment_parameter.setId(assignment_id):
        raise RuntimeError(f'Unable to generate assignment with id '
                           f'{assignment_id}. Invalid SBML identifier.')

    assignment_parameter.setName(assignment_id)
    assignment_parameter.setConstant(False)
    assignment_parameter.setUnits('dimensionless')

    assignment_rule = model.createAssignmentRule()
    assignment_rule.setVariable(assignment_id)

    sbml_formula = sbml.parseL3Formula(formula)

    if sbml_formula is not None:
        assignment_rule.setMath(sbml.parseL3Formula(formula))
    else:
        raise RuntimeError(f'Unable to generate assignment for formula '
                           f'{formula}, libsbml can not parse the given '
                           f'expression.')


def _read_functions_block(model: sbml.Model, functions_list: list):
    """
    Read and process the functions block in the YAML file.

    The expected format of a function definition is:
        {'functionId': <functionId>,
         'arguments': <arguments>,
         'formula' : <formula>}

    Arguments:
        model: a SBML model
        functions_list: a list of dictionaries where each entry is a
                        function definition
    """
    for function_def in functions_list:
        _create_function(model,
                         function_def['functionId'],
                         function_def['arguments'],
                         function_def['formula'])


def _create_function(model: sbml.Model,
                     function_id: str,
                     arguments: str,
                     formula: str):
    """
    Create a function definition and add it to the SBML model.

    Arguments:
        model: SBML model to which the function will be added.
        function_id: the function id/name
        arguments: the arguments of the function (species AND parameters)
        formula: the formula of the function
    """
    f = model.createFunctionDefinition()

    if f.setId(function_id) != sbml.LIBSBML_OPERATION_SUCCESS:
        raise RuntimeError(f'Unable to generate function with id '
                           f'{function_id}. Invalid SBML identifier.')

    math = sbml.parseL3Formula('lambda(' + arguments + ', ' + formula + ')')

    if math is not None:
        f.setMath(math)
    else:
        raise RuntimeError(f'Unable to generate assignment for funtion '
                           f'{function_id}, libsbml can not parse the given '
                           f'function expression, given by '
                           f'lambda({arguments} , {formula}).')


def _read_odes_block(model: sbml.Model, odes_list: list):
    """
    Read and process the odes block in the YAML file.

    In particular, read the ODEs and add a species for the corresponding
    state and the right hand side as rateRules to the given SBML file.

    The expected format of an ode definition is:
    {'stateId': <state_variable>, 'rightHandSide' : <right_hand_side>,
    'initialValue' = <initial_value>}

    Arguments:
        model: a SBML model
        odes_list: block of ODE definitions
    """
    for ode_def in odes_list:
        _create_species(model, ode_def['stateId'], ode_def['initialValue'])
        _create_rate_rule(model, ode_def['stateId'], ode_def['rightHandSide'])


def _create_species(model: sbml.Model, species_id: str, initial_amount: str):
    """
    Create a species and add it to the SBML model.

    Units are set as dimensionless by default.

    Arguments:
        model: the SBML model to which the species will be added.
        species_id: the species ID
        initial_amount: the species initial amount

    Returns:
        s: the SBML species
    """
    s = model.createSpecies()
    if s.setId(species_id) != sbml.LIBSBML_OPERATION_SUCCESS:
        raise RuntimeError(f'Unable to generate species with id '
                           f'{species_id}. Invalid SBML identifier.')

    try:
        s.setInitialAmount(float(initial_amount))
    except ValueError:
        init = model.createInitialAssignment()
        init.setId('init_' + species_id)
        init.setSymbol(species_id)
        init.setMath(sbml.parseL3Formula(initial_amount))

    s.setConstant(False)
    s.setBoundaryCondition(False)
    s.setHasOnlySubstanceUnits(False)
    s.setCompartment('Compartment')

    s.setSubstanceUnits('dimensionless')

    return s


def _create_rate_rule(model: sbml.Model, species_id: str, formula: str):
    """
    Create an SBML rateRule for a species and add it to the SBML model.

    This is where the ODEs are encoded.

    Arguments:
        model: SBML model to which the rate rule will be added.
        species_id: the id of the state of the ODE
        formula: the right-hand-side of the ODE
    """
    r = model.createRateRule()
    r.setId('d_dt_' + species_id)
    r.setVariable(species_id)
    math_ast = sbml.parseL3Formula(formula)
    if math_ast is None:
        raise RuntimeError(f'Unable to generate the rate rule for the state '
                           f'{species_id}, libsbml can not parse the right-'
                           f'hand side, given by {formula}).')
    r.setMath(math_ast)


def _read_observables_block(model: sbml.Model,
                            observable_list: list,
                            observables_as_assignments: bool):
    """
    Read and process the observables block in the YAML file.

    Since the observables are not represented in the SBML, it only gives
    a warning to inform the user.

    Arguments:
        model: SBML model (libsbml)
        observable_list: observables block containing all
                         observable definitions.
        observables_as_assignments: indicates whether there should be
            parameter assignments of the form `observable_<observable_id>`.
    """
    if observables_as_assignments:
        for observable_def in observable_list:
            _create_assignment(model,
                               f'observable_{observable_def["observableId"]}',
                               observable_def['observableFormula'])
    else:
        warnings.warn(
            'Observables are not represented in the SBML and therefore only '
            'have an effect on the output when called via yaml2petab')


def _read_conditions_block(model: sbml.Model, conditions_list: list):
    """
    Read and process the conditions block in the YAML file.

    Since conditions are not represented in the SBML, it only gives
    a warning to inform the user.

    Arguments:
        model: SBML model (libsbml)
        conditions_list: conditions block containing all
                         conditions definitions.
    """
    warnings.warn(
        'Conditions are not represented in the SBML and therefore only have '
        'an effect on the output when called via yaml2petab')


def main():
    """Command-Line Interface."""
    parser = argparse.ArgumentParser(
        description='Takes in an ODE model in .yaml and converts it to SBML.')
    parser.add_argument('yaml_file', type=str, help='Path to input YAML file.')
    parser.add_argument('sbml_file', type=str,
                        help='Path to output SBML file.')
    parser.add_argument('-o', '--observables_as_assignments',
                        action='store_true',
                        help='Optional argument, flag, which indicates, if '
                             'observables should be represented in the SBML'
                             'as assignments. Potential Values: 1/0 (yes/no).')

    args = parser.parse_args()

    print(f'Path to YAML file: {args.yaml_file}')
    print(f'Path to SBML file: {args.sbml_file}')

    print('Converting...')

    yaml2sbml(args.yaml_file,
              args.sbml_file,
              args.observables_as_assignments)


if __name__ == '__main__':
    main()



import os
import unittest


from yaml2sbml.yaml_validation import validate_yaml
from jsonschema.exceptions import ValidationError


class TestYamlValidation(unittest.TestCase):

    def setUp(self):
        this_dir, _ = os.path.split(__file__)
        self.test_folder = os.path.join(this_dir, 'test_yaml2sbml')

    def test_validate_yaml_valid_1(self):
        # Test validation for a valid yaml model.
        file_in = os.path.join(self.test_folder, 'ode_input1.yaml')
        validate_yaml(file_in)

    def test_validate_yaml_valid_2(self):
        # Test validation for a second valid yaml model.
        file_in = os.path.join(self.test_folder, 'ode_input2.yaml')
        validate_yaml(file_in)

    def test_validate_yaml_typos(self):
        # Test if Error is thrown due to typos.
        file_in = os.path.join(self.test_folder, 'ode_input_typos.yaml')
        with self.assertRaises(ValidationError):
            validate_yaml(file_in)

    def test_validate_yaml_typos_required(self):
        # Test if Error is thrown due to typos.
        file_in = os.path.join(self.test_folder,
                               'ode_input_typos_required.yaml')
        with self.assertRaises(ValidationError):
            validate_yaml(file_in)

    def test_validate_yaml_empty_section(self):
        # Test if Error is thrown due to empty observable section.
        file_in = os.path.join(self.test_folder,
                               'ode_input_empty_section.yaml')
        with self.assertRaises(ValidationError):
            validate_yaml(file_in)

    def test_catch_invalid_time_block_missing_variable_key(self):
        # Test time block without kew word "variable".
        file_in = os.path.join(self.test_folder,
                               'ode_input_invalid_time_1.yaml')
        with self.assertRaises(ValidationError):
            validate_yaml(file_in)

    def test_catch_invalid_time_block_as_array(self):
        # Test time block as array instead of single object.
        file_in = os.path.join(self.test_folder,
                               'ode_input_invalid_time_2.yaml')
        with self.assertRaises(ValidationError):
            validate_yaml(file_in)


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(TestYamlValidation())
    unittest.main()



"""Import of yaml2sbml's public API."""

# version
from .version import __version__   # noqa: F401

# API
from .yaml2sbml import yaml2sbml
from .yaml2PEtab import yaml2petab, validate_petab_tables
from .yaml_validation import validate_yaml
from .YamlModel import YamlModel



"""Validator of the input yaml."""
import os
import yaml
import jsonschema
from yaml.scanner import ScannerError
import argparse


SCHEMA = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                      "yaml_schema.yaml")


def validate_yaml(yaml_dir: str):
    """
    Validate the syntax of the YAML file.

    Arguments:
        yaml_dir: path to YAML file to be validated

    Returns:
        jsonschema.validate
    """
    try:

        with open(yaml_dir, 'r') as f_in:
            yaml_contents = f_in.read()
            yaml_dict = yaml.full_load(yaml_contents)

    except ScannerError:
        raise RuntimeError('YAML file can not be parsed due to a Scanner '
                           'Error. This commonly happens if formulas begin '
                           'with a minus. Please set them inside of brackets '
                           '"(...)" or quotation marks.')

    _validate_yaml_from_dict(yaml_dict)
    print('YAML file is valid ✅')


def _validate_yaml_from_dict(yaml_dict: dict):
    """
    Validate the syntax of the YAML file, using a dict as input.

    Arguments:
        yaml_dict: YAML model as dict.

    Returns:
        jsonschema.validate
    """
    # read in SCHEMA
    with open(SCHEMA, 'r') as f_in:
        yaml_contents = f_in.read()
        schema = yaml.full_load(yaml_contents)

    jsonschema.validate(instance=yaml_dict, schema=schema)


def main():
    """Command-Line Interface."""
    parser = argparse.ArgumentParser(
        description='Validates a yaml model '
                    'so that it can be used by yaml2bsml.')

    parser.add_argument('yaml_file', type=str,
                        help='Directory of yaml file, that '
                             'should be validated.')

    args = parser.parse_args()

    print(f'Path to yaml file: {args.yaml_file}')
    print('Validating...')

    validate_yaml(args.yaml_file)


if __name__ == '__main__':
    main()



import os
import unittest

from yaml2sbml.yaml2sbml import yaml2sbml, _parse_yaml


class TestYaml2SBML(unittest.TestCase):
    """
    TestCase class for testing ODE import from a generic yaml file
    and conversion to SBML.
    """

    def setUp(self):
        this_dir, _ = os.path.split(__file__)
        self.test_folder = os.path.join(this_dir, 'test_yaml2sbml')

    def test_yaml_import(self):
        """
        Test yaml import/SBML generation...
        """
        yaml_dir = os.path.join(self.test_folder, 'ode_input1.yaml')
        expected_result_file = os.path.join(self.test_folder,
                                            'true_sbml_output.xml')

        sbml_test_dir = os.path.join(self.test_folder, 'sbml_test.xml')

        yaml2sbml(yaml_dir, sbml_test_dir)

        with open(expected_result_file, 'r') as f_in:
            expected_sbml = f_in.read()

        with open(sbml_test_dir, 'r') as f_out:
            observed_sbml = f_out.read()

        # check if lines coincide
        for line_true, line_tested in zip(expected_sbml.split('\n'),
                                          observed_sbml.split('\n')):

            # the line containing name & id will not match...
            if not line_true.startswith('  <model id='):
                self.assertEqual(line_true, line_tested)

        os.remove(sbml_test_dir)

    def test_yaml_import_observables(self):
        """
        Test yaml import/export for a model containing observables.
        """
        yaml_dir = os.path.join(self.test_folder, 'ode_input2.yaml')

        expected_result_file = \
            os.path.join(self.test_folder, 'true_sbml_output.xml')

        sbml_test_dir = os.path.join(self.test_folder, 'sbml_test.xml')

        # Call yaml2sbml with observables as assignments
        yaml2sbml(yaml_dir,
                  sbml_test_dir,
                  observables_as_assignments=True)

        # Call yaml2sbml with observables not translated
        yaml2sbml(yaml_dir, sbml_test_dir)

        with open(expected_result_file, 'r') as f_in:
            expected_sbml = f_in.read()

        with open(sbml_test_dir, 'r') as f_out:
            observed_sbml = f_out.read()

        for line_true, line_tested in zip(expected_sbml.split('\n'),
                                          observed_sbml.split('\n')):

            # the line containing name & id will not match...
            if not line_true.startswith('  <model id='):
                self.assertEqual(line_true, line_tested)

        os.remove(sbml_test_dir)

    def test_catch_invalid_sbml_identifier(self):
        """
        Check for invalid SBML identifiers.
        """
        yaml_dir = os.path.join(self.test_folder,
                                'ode_input_invalid_SBML_identifier.yaml')
        with self.assertRaises(RuntimeError):
            _parse_yaml(yaml_dir,
                        'Test_Model')

    def test_catch_invalid_math(self):
        """
        Check for strings that parseL3Formula can not parse.
        """
        yaml_dir = os.path.join(self.test_folder,
                                'ode_input_invalid_formula.yaml')
        with self.assertRaises(RuntimeError):
            _parse_yaml(yaml_dir,
                        'Test_Model')


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(TestYaml2SBML())
    unittest.main()



"""A Model Editor for creating YAML models."""
import yaml
import os.path
import copy
from typing import Union
from pathlib import Path

from .yaml2sbml import _parse_yaml_dict, _load_yaml_file
from .yaml2PEtab import _yaml2petab
from .yaml_validation import _validate_yaml_from_dict


class YamlModel:
    """Functionality to set up, edit, load and write yaml models."""

    def __init__(self):
        """Initialize a YAML model."""
        self._yaml_model = {'time': {},
                            'odes': [],
                            'parameters': [],
                            'assignments': [],
                            'functions': [],
                            'observables': [],
                            'conditions': []}

    @staticmethod
    def load_from_yaml(yaml_dir: str):
        """
        Create a model instance from a YAML file.

        Arguments:
            yaml_dir:
                directory to the YAML file, that should be imported

        Returns:
            cls:
                new model
        """
        new_model = YamlModel()

        yaml_contents = _load_yaml_file(yaml_dir)
        new_model._yaml_model.update(yaml_contents)

        # check, if the model is valid
        new_model.validate_model()

        return new_model

    def write_to_yaml(self,
                      yaml_dir: str,
                      overwrite: bool = False):
        """
        Write the model to a YAML file given as `yaml_dir`.

        Arguments:
            yaml_dir:
                path/file, where the YAML should be written
            overwrite:
                Indicates, whether an existing YAML should be overwritten

        Raises:
            ValueError
            FileExistsError
        """
        if not (yaml_dir.endswith('.yaml') or yaml_dir.endswith('.yml')):
            raise ValueError('yaml_dir should contain path to the yaml '
                             'and hence end with .yaml or .yml')

        if (not overwrite) and os.path.exists(yaml_dir):
            raise FileExistsError(f'Can not write yaml model. File {yaml_dir}'
                                  f' already exists. Consider to set '
                                  f'overwrite=True.')

        reduced_model_dict = self._get_reduced_model_dict()

        # translate to string
        yaml_as_string = yaml.dump(reduced_model_dict,
                                   sort_keys=False,
                                   indent=6)

        # post-process: add empty line around blocks
        for key in self._yaml_model.keys():
            yaml_as_string = yaml_as_string.replace(f'{key}:',
                                                    f'\n{key}:')
        yaml_as_string = yaml_as_string.replace('-     ', '\n    - ')

        with open(yaml_dir, 'w') as file:
            file.write(yaml_as_string)

    def write_to_sbml(self,
                      sbml_dir: str,
                      overwrite: bool = False):
        """
        Write the model as an SBML file to the directory given in `sbml_dir`.

        Arguments:
            sbml_dir:
                path/file, where the sbml should be written
            overwrite:
                Indicates, whether an existing yaml should be overwritten

        Raises:
            ValueError
            FileExistsError
        """
        # Check file ending.
        if not (sbml_dir.endswith('.xml') or sbml_dir.endswith('.sbml')):
            raise ValueError('sbml_dir should contain path to the sbml '
                             'and hence end with .xml or .sbml')

        # model name = sbml name without file extension
        model_name = Path(sbml_dir).stem

        if (not overwrite) and os.path.exists(sbml_dir):
            raise FileExistsError(f'Can not write SBML model. File {sbml_dir}'
                                  f' already exists. Consider to set '
                                  f'overwrite=True.')

        # generate SBML as string
        reduced_model_dict = self._get_reduced_model_dict()
        sbml_as_string = _parse_yaml_dict(reduced_model_dict,
                                          model_name)

        with open(sbml_dir, 'w') as f_out:
            f_out.write(sbml_as_string)

    def write_to_petab(self,
                       output_dir: str,
                       model_name: str,
                       petab_yaml_name: str = None,
                       measurement_table_name: str = None):
        """
        Write the YamlModel as a PEtab problem.

        Equivalent to calling `yaml2petab on the file produced by the YAML
        output.

        If a `petab_yaml_name is given, a YAML file is created, that organizes
        the petab problem. If additionally a `measurement_table_file_name is
        specified, this file name is written into the created YAML file.

        Arguments:
            output_dir: path the output file(s) are be written out
            model_name: name of SBML model
            petab_yaml_name: name of the YAML organizing the PEtab problem.
            measurement_table_name: Name of measurement table
        """
        reduced_model_dict = self._get_reduced_model_dict()

        _yaml2petab(reduced_model_dict,
                    output_dir,
                    model_name,
                    petab_yaml_name,
                    measurement_table_name)

    def validate_model(self):
        """
        Validate the YAML model.

        Raises:
            ValidationError
        """
        _validate_yaml_from_dict(self._get_reduced_model_dict())

    def _get_reduced_model_dict(self) -> dict:
        """
        Return a reduced model dict, where keys without an entry are deleted.

        Return a copy of the model dict!!

        Returns:
            reduced_model_dict
        """
        reduced_model_dict = {}

        for (key, val) in self._yaml_model.items():
            if val:
                reduced_model_dict[key] = copy.deepcopy(val)

        return reduced_model_dict

    # functionalities regarding the time
    def is_set_time(self):
        """Check whether there is a time variable."""
        return 'variable' in self._yaml_model['time'].keys()

    def set_time(self,
                 time_variable: str):
        """Set time variable."""
        self._yaml_model['time'] = {'variable': time_variable}

    def delete_time(self):
        """Delete time variable."""
        self._yaml_model['time'] = {}

    def get_time(self):
        """Get time variable."""
        if self.is_set_time():
            return self._yaml_model['time']['variable']
        else:
            return None

    # functions adding a value
    def add_parameter(self,
                      parameter_id: str,
                      overwrite: bool = False,
                      nominal_value: float = None,
                      parameter_name: str = None,
                      parameter_scale: str = None,
                      lower_bound: float = None,
                      upper_bound: float = None,
                      estimate: int = None):
        """
        Add a parameter.

        Overwrite an existing parameter with the same id, if `overwrite==True`.

        Arguments:
            parameter_id:
                str, parameter id
            overwrite:
                bool, indicates if an existing state/ODE should be overwritten
            nominal_value:
                float, nominal value of the parameter.
            parameter_name:
                str, name of parameter in PEtab parameter table, optional.
            parameter_scale
                str. scale of parameter in PEtab parameter table, optional.
            lower_bound:
                float, lower bound of parameter in PEtab parameter table,
                optional.
            upper_bound:
                float, upper bound of parameter in PEtab parameter table,
                optional.
            estimate:
                int, estimate flag of parameter in PEtab parameter table,
                optional.
        """
        # if parameter exists: delete if overwrite
        if parameter_id in self.get_parameter_ids():
            if overwrite:
                self.delete_parameter(parameter_id)
            else:
                raise ValueError('Could not add parameter with id'
                                 f' {parameter_id}: Parameter with the same'
                                 ' id already exists.')

        entry_dict = {'parameterId': parameter_id,
                      'nominalValue': nominal_value,
                      'parameterName': parameter_name,
                      'parameterScale': parameter_scale,
                      'lowerBound': lower_bound,
                      'upperBound': upper_bound,
                      'estimate': estimate}

        self._add_entry(entry_dict, 'parameters')

    def add_ode(self,
                state_id: str,
                right_hand_side: Union[float, str],
                initial_value: Union[float, str],
                overwrite: bool = False):
        """
        Add state/ODE.

        Overwrite an existing state/ODE with the same id, if `overwrite==True`.

        Arguments:
            state_id:
                str, state id
            right_hand_side:
                str or float, right hand side of the ODE.
            initial_value:
                str or float, initial value of the ODE at t=0
            overwrite:
                bool, indicates if an existing state/ODE should be overwritten
        """
        # if state exists: delete if overwrite
        if state_id in self.get_ode_ids():
            if overwrite:
                self.delete_ode(state_id)
            else:
                raise ValueError(f'Could not add state/ODE with id {state_id}:'
                                 f' State with the same id already exists.')

        entry_dict = {'stateId': state_id,
                      'rightHandSide': right_hand_side,
                      'initialValue': initial_value}

        self._add_entry(entry_dict, 'odes')

    def add_assignment(self,
                       assignment_id: str,
                       formula: str,
                       overwrite: bool = False):
        """
        Add assignment.

        Overwrite an existing assignment with the same id, if
        `overwrite==True`.

        Arguments:
            assignment_id:
                str, function id
            formula:
                str, right hand side of assignment definition.
            overwrite:
                bool, indicates if an existing assignment should be overwritten
        """
        # if assignment exists: delete if overwrite
        if assignment_id in self.get_assignment_ids():
            if overwrite:
                self.delete_assignment(assignment_id)
            else:
                raise ValueError('Could not add assignment with id '
                                 f'{assignment_id}: Assignment with the same '
                                 'id already exists.')

        entry_dict = {'assignmentId': assignment_id,
                      'formula': formula}

        self._add_entry(entry_dict, 'assignments')

    def add_function(self,
                     function_id: str,
                     arguments: str,
                     formula: str,
                     overwrite: bool = False):
        """
        Add function.

        Overwrite an existing function with the same id, if `overwrite==True`.

        Arguments:
            function_id:
                str, function id
            arguments:
                str, arguments, separated by a comma
            formula:
                str, right hand side of the function definition
            overwrite:
                bool, indicates if an existing function should be overwritten
        """
        # if function exists: delete if overwrite
        if function_id in self.get_function_ids():
            if overwrite:
                self.delete_function(function_id)
            else:
                raise ValueError('Could not add function with id '
                                 f' {function_id}: Function with the same '
                                 'id already exists.')

        entry_dict = {'functionId': function_id,
                      'arguments': arguments,
                      'formula': formula}

        self._add_entry(entry_dict, 'functions')

    def add_observable(self,
                       observable_id: str,
                       observable_formula: str,
                       noise_formula: str,
                       overwrite: bool = False,
                       observable_name: str = None,
                       observable_transformation: str = None,
                       noise_distribution: str = None):
        """
        Add observable.

        Observables are not represented inside an SBML and only play a role
        when generating a PEtab problem see PEtabs observable table).

        Overwrite an existing observable with the same id, if
        `overwrite==True`.

        Arguments:
            observable_id:
                str, observable id
            observable_formula:
                str, formula of the observable function
            noise_formula:
                str, formula of the noise
            overwrite:
                bool, indicates if an existing observable should be overwritten
            observable_name:
                Observable name. Optional.
            observable_transformation:
                Observable transformation ('lin'/'log'/'log10'). Optional
        """
        # if observable exists: delete if overwrite
        if observable_id in self.get_observable_ids():
            if overwrite:
                self.delete_observable(observable_id)
            else:
                raise ValueError('Could not add observable with id '
                                 f'{observable_id}: Observable with the same '
                                 'id already exists.')

        entry_dict = {'observableId': observable_id,
                      'observableName': observable_name,
                      'observableFormula': observable_formula,
                      'observableTransformation': observable_transformation,
                      'noiseFormula': noise_formula,
                      'noiseDistribution': noise_distribution}

        self._add_entry(entry_dict, 'observables')

    def add_condition(self,
                      condition_id: str,
                      condition_dict: dict,
                      overwrite: bool = False,
                      condition_name: str = None):
        """
        Add condition `condition_id`.

        Conditions are not represented inside an SBML and only play a role
        when generating a PEtab problem (see PEtabs condition table).

        Overwrite an existing condition with the same id, if `overwrite==True`.

        Arguments:
            condition_id:
                str, condition id
            condition_dict:
                dict, of the form {<parameter or state id>: <value>}.
                Corresponds to entries in the PEtab condition table.
                See details there.
            overwrite:
                bool, indicates if an existing condition should be overwritten
            condition_name:
                Condition name. Optional.
        """
        # if condition exists: delete if overwrite
        if condition_id in self.get_condition_ids():
            if overwrite:
                self.delete_condition(condition_id)
            else:
                raise ValueError('Could not add condition with id '
                                 f' {condition_id}: Condition with the same '
                                 f'id already exists.')

        entry_dict = {'conditionId': condition_id,
                      'conditionName': condition_name,
                      **condition_dict}

        self._add_entry(entry_dict, 'conditions')

    def _add_entry(self,
                   entry_dict: dict,
                   block_key: str):
        """
        Add the entry in 'entry_dict' to the block indexed by 'block_key'.

        If 'overwrite=True', an existing value for that key is overwritten.

        Arguments:
        entry_dict:
            dict, that stores the new entry
        block_key:
            name, where the ids should be searched (e.g. 'parameters')
        """
        # filter out None values and append
        filtered_dict = _filter_none_values(entry_dict)
        self._yaml_model[block_key].append(filtered_dict)

    # functionalities to get ids
    def get_parameter_ids(self):
        """Return a list with all parameter ids."""
        return self._get_ids('parameters', 'parameterId')

    def get_ode_ids(self):
        """Return a list with all state ids."""
        return self._get_ids('odes', 'stateId')

    def get_assignment_ids(self):
        """Return a list with all assignment ids."""
        return self._get_ids('assignments', 'assignmentId')

    def get_function_ids(self):
        """Return a list with all function ids."""
        return self._get_ids('functions', 'functionId')

    def get_observable_ids(self):
        """Return a list with all observable ids."""
        return self._get_ids('observables', 'observableId')

    def get_condition_ids(self):
        """Return a list with all conditions ids."""
        return self._get_ids('conditions', 'conditionId')

    def _get_ids(self,
                 block_key: str,
                 index_key: str):
        """
        Return all ids in the corresponding block.

        Arguments:
        block_key:
            name, where the ids should be searched (e.g. 'parameters')
        index_key:
            key of the identifier, in that block (e.g. 'parameterId')

        Returns:
        res:
            list of ids
        """
        return [element[index_key] for element in self._yaml_model[block_key]]

    # functionalities to get entry by Id:
    def get_parameter_by_id(self,
                            parameter_id: str):
        """
        Return dict for corresponding parameter.

        Raise a `ValueError, if the parameter does not exist.
        """
        if parameter_id not in self.get_parameter_ids():
            raise IndexError(f'Could not find parameter {parameter_id}.')

        return self._get_entry_by_id('parameters',
                                     'parameterId',
                                     parameter_id)

    def get_ode_by_id(self,
                      state_id: str):
        """
        Return dict for corresponding ODE/state.

        Raise a `ValueError`, if the ODE/state does not exist.
        """
        if state_id not in self.get_ode_ids():
            raise IndexError(f'Could not find state/ODE {state_id}.')

        return self._get_entry_by_id('odes',
                                     'stateId',
                                     state_id)

    def get_assignment_by_id(self,
                             assignment_id: str):
        """
        Return dict for corresponding assignment.

        Raise a `ValueError`, if the assignment does not exist.
        """
        if assignment_id not in self.get_assignment_ids():
            raise IndexError(f'Could not find assignment {assignment_id}.')

        return self._get_entry_by_id('assignments',
                                     'assignmentId',
                                     assignment_id)

    def get_function_by_id(self,
                           function_id: str):
        """
        Return dict for corresponding function.

        Raise a `ValueError`, if the function does not exist.
        """
        if function_id not in self.get_function_ids():
            raise IndexError(f'Could not find function {function_id}.')

        return self._get_entry_by_id('functions',
                                     'functionId',
                                     function_id)

    def get_observable_by_id(self,
                             observable_id: str):
        """
        Return dict for corresponding observable.

        Raise a `ValueError`, if the observable does not exist.
        """
        if observable_id not in self.get_observable_ids():
            raise IndexError(f'Could not find observable {observable_id}.')

        return self._get_entry_by_id('observables',
                                     'observableId',
                                     observable_id)

    def get_condition_by_id(self,
                            condition_id: str):
        """
        Return dict for corresponding condition.

        Raise a `ValueError`, if the condition does not exist.
        """
        if condition_id not in self.get_condition_ids():
            raise IndexError(f'Could not find condition {condition_id}.')

        return self._get_entry_by_id('conditions',
                                     'conditionId',
                                     condition_id)

    def _get_entry_by_id(self,
                         block_key: str,
                         index_key: str,
                         entry_id: str):
        """
        Get entry by id in corresponding block.

        Returns the entry with id 'entry_id' in the block named
        `block_key` in self._yaml_model. 'index_key' gives the name of the
        id-key.

        Arguments:
        block_key:
            Key of the block (e.g. 'parameters')
        index_key:
            key of the identifier in that block (e.g. 'parameterId')
        entry_id:
            Id of element, that should be returned

        Returns:
            Entry_dict. None, if entry is not found.
        """
        # check if block exists.
        for entry in self._yaml_model[block_key]:
            if entry[index_key] == entry_id:
                return entry

        return None

    # functionalities to delete entry by Id:
    def delete_parameter(self,
                         parameter_id: str):
        """
        Delete a parameter.

        Raise a ValueError, if parameter does not exist.
        """
        if not self._delete_entry('parameters',
                                  'parameterId',
                                  parameter_id):

            raise ValueError(f'Could not delete parameter {parameter_id}. '
                             f'Invalid parameterId.')

    def delete_ode(self,
                   state_id: str):
        """
        Delete a state + ODE.

        Raise a ValueError, if state does not exist.
        """
        if not self._delete_entry('odes',
                                  'stateId',
                                  state_id):

            raise ValueError(f'Could not delete ODE for state {state_id}. '
                             f'Invalid stateId.')

    def delete_assignment(self,
                          assignment_id: str):
        """
        Delete an assignment.

        Raise a ValueError, if assignment does not exist.
        """
        if not self._delete_entry('assignments',
                                  'assignmentId',
                                  assignment_id):

            raise ValueError(f'Could not delete assignment {assignment_id}. '
                             f'Invalid assignmentId.')

    def delete_function(self,
                        function_id: str):
        """
        Delete a function.

        Raise a ValueError, if function does not exist.
        """
        if not self._delete_entry('functions',
                                  'functionId',
                                  function_id):

            raise ValueError(f'Could not delete function {function_id}. '
                             f'Invalid functionId.')

    def delete_observable(self,
                          observable_id: str):
        """
        Delete an observable.

        Raise a ValueError, if observable does not exist.
        """
        if not self._delete_entry('observables',
                                  'observableId',
                                  observable_id):

            raise ValueError(f'Could not delete observable {observable_id}. '
                             f'Invalid observableId.')

    def delete_condition(self,
                         condition_id: str):
        """
        Delete a condition.

        Raise a ValueError, if condition does not exist.
        """
        if not self._delete_entry('conditions',
                                  'conditionId',
                                  condition_id):

            raise ValueError(f'Could not delete condition {condition_id}. '
                             f'Invalid conditionId.')

    def _delete_entry(self,
                      block_key: str,
                      index_key: str,
                      deleted_object_id: str):
        """
        Delete entry in block.

        Delete the entry with id 'deleted_object_id' in the block named
        `block_key` in self._yaml_model. 'index_key' gives the name of the
        id-key.

        Arguments:
        block_key:
            Key of the block (e.g. 'parameters')
        index_key:
            key of the identifier in that block (e.g. 'parameterId')
        deleted_object_id:
            Id of element, that should be deleted

        Returns:
            Bool, that indicates, whether deletion was successful.
        """
        # Check if block exists
        for i, entry in enumerate(self._yaml_model[block_key]):

            # search for entry and delete
            if entry[index_key] == deleted_object_id:
                self._yaml_model[block_key].pop(i)
                return True

        return False


def _filter_none_values(d: dict):
    """
    Filter out the key-value pairs with `None` as value.

    Arguments:
    d
        dictionary

    Returns:
        filtered dictionary.
    """
    return {key: value for (key, value) in d.items() if value is not None}



import os
import shutil
import unittest

import yaml2sbml.yaml2PEtab as yaml2PEtab


class TestYaml2PEtab(unittest.TestCase):
    """
    TestCase class for testing ODE import from a generic yaml file
    and conversion to PEtab.
    """

    def setUp(self):
        this_dir, _ = os.path.split(__file__)
        self.input_folder = os.path.join(this_dir, 'test_yaml2sbml')
        self.output_folder = os.path.join(this_dir, 'test_yaml2sbml_output')

    def tearDown(self):
        shutil.rmtree(self.output_folder)

    def test_petab_export(self):
        """
        Test PEtab export
        """
        input_yaml_dir = os.path.join(self.input_folder, 'ode_input2.yaml')

        yaml2PEtab.yaml2petab(input_yaml_dir,
                              self.output_folder,
                              'sbml_test.xml')

        yaml2PEtab.validate_petab_tables(
            os.path.join(self.output_folder, 'sbml_test.xml'),
            self.output_folder)


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(TestYaml2PEtab())
    unittest.main()



import os
import shutil

from yaml2sbml.yaml2sbml import _parse_yaml


def test_yaml2sbml_cli(script_runner):
    """Test the command line command `yaml2sbml`."""
    path = os.path.dirname(os.path.abspath(__file__))

    yaml_dir = os.path.join(path, 'test_yaml2sbml', 'ode_input1.yaml')
    sbml_dir = os.path.join(path, 'test_sbml.xml')

    # run CLI command
    script_runner.run('yaml2sbml', yaml_dir, sbml_dir)

    # test if SBML created by the python yaml2sbml command is the same.
    sbml_from_python = _parse_yaml(yaml_dir, 'test_sbml')

    with open(sbml_dir, 'r') as f_in:
        sbml_from_cli = f_in.read()

    for line_cli, line_python in zip(sbml_from_cli.split('\n'),
                                     sbml_from_python.split('\n')):

        # the line containing name & id will not match...
        if not line_cli == line_python:
            raise AssertionError('SBML from CLI and '
                                 'Python interface differ.')

    os.remove(sbml_dir)


def test_yaml2petab_cli(script_runner):
    """Test the command line command `yaml2petab."""
    path = os.path.dirname(os.path.abspath(__file__))
    yaml_dir = os.path.join(path, 'test_yaml2sbml', 'ode_input2.yaml')
    output_dir = os.path.join(path, 'test_output')
    model_name = 'petab_test_sbml.xml'

    # run with no optional Arguments
    script_runner.run('yaml2petab', yaml_dir, output_dir, model_name)

    # run with optional Arguments
    script_runner.run('yaml2petab',
                      yaml_dir,
                      output_dir,
                      model_name,
                      '-y test_yaml.yml')

    # delete the generated files
    shutil.rmtree(output_dir)


def test_yaml2sbml_validate_cli(script_runner):
    """Test the command line command ``yaml2sbml_validate`"""

    path = os.path.dirname(os.path.abspath(__file__))
    yaml_dir = os.path.join(path, 'test_yaml2sbml/ode_input1.yaml')

    script_runner.run('yaml2sbml_validate', yaml_dir)



import os
import shutil

import unittest
import tempfile

from yaml2sbml.YamlModel import YamlModel
from yaml2sbml.yaml_validation import validate_yaml


class TestYamlModel(unittest.TestCase):
    """TestCase class for testing YamlModel."""

    def setUp(self):
        # input directory
        this_dir, _ = os.path.split(__file__)
        self.test_input_folder = os.path.join(this_dir, 'test_yaml2sbml')

        # temporary output directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # remove temporary directory
        shutil.rmtree(self.test_dir)

    def test_load_and_write(self):
        """
        Test loading and writing.
        """
        # read in yaml
        yaml_dir = os.path.join(self.test_input_folder, 'ode_input2.yaml')
        model = YamlModel.load_from_yaml(yaml_dir)

        # save and read in again
        saved_yaml_file = os.path.join(self.test_dir, 'test_yaml.yml')
        model.write_to_yaml(saved_yaml_file)
        reloaded_model = YamlModel.load_from_yaml(saved_yaml_file)

        # check if save+reload did change the model
        self.assertDictEqual(model._yaml_model, reloaded_model._yaml_model)

    def test_time(self):
        """
        Test all functionality regarding the time keyword.
        """
        model = YamlModel()
        time_var = 't'

        self.assertFalse(model.is_set_time())

        model.set_time(time_var)

        self.assertTrue(model.is_set_time())
        self.assertEqual(model.get_time(), time_var)

        model.delete_time()
        self.assertFalse(model.is_set_time())

    def test_parameter(self):
        """
        Test all functionality regarding the 'parameters' keyword.
        """
        model = YamlModel()
        parameter_id = 'p1'
        model.add_parameter(parameter_id,
                            nominal_value=0,
                            lower_bound=-2,
                            upper_bound=2)

        # test get_parameter_by_id
        self.assertIsInstance(model.get_parameter_by_id(parameter_id), dict)

        # test overwrite
        with self.assertRaises(ValueError):
            model.add_parameter(parameter_id=parameter_id)

        # now over write parameter
        model.add_parameter(parameter_id=parameter_id,
                            overwrite=True)

        # test get_parameter_ids
        self.assertListEqual(model.get_parameter_ids(),
                             [parameter_id])

        # test delete
        model.delete_parameter(parameter_id)
        self.assertListEqual(model.get_parameter_ids(), [])

    def test_ode(self):
        """
        Test all functionality regarding the 'odes' keyword.
        """
        model = YamlModel()

        state_id = 's1'
        right_hand_side = '1'
        initial_value = 0

        model.add_ode(state_id, right_hand_side, initial_value)

        # test get_ode_by_id
        self.assertIsInstance(model.get_ode_by_id(state_id), dict)

        # test overwrite
        with self.assertRaises(ValueError):
            model.add_ode(state_id, right_hand_side, initial_value)

        # now over write ode/state
        model.add_ode(state_id,
                      right_hand_side,
                      initial_value,
                      overwrite=True)

        # test get_ode_ids
        self.assertListEqual(model.get_ode_ids(),
                             [state_id])

        # test delete
        model.delete_ode(state_id)
        self.assertListEqual(model.get_ode_ids(), [])

    def test_assignment(self):
        """
        Test all functionality regarding the 'assignments' keyword.
        """
        model = YamlModel()

        assignment_id = 's1'
        formula = '1+2'

        model.add_assignment(assignment_id, formula)

        # test get_assignment_by_id
        self.assertIsInstance(model.get_assignment_by_id(assignment_id), dict)

        # test overwrite
        with self.assertRaises(ValueError):
            model.add_assignment(assignment_id, formula)

        # now over write assignment
        model.add_assignment(assignment_id, formula, overwrite=True)

        # test get_assignment_ids
        self.assertListEqual(model.get_assignment_ids(),
                             [assignment_id])

        # test delete
        model.delete_assignment(assignment_id)
        self.assertListEqual(model.get_assignment_ids(), [])

    def test_function(self):
        """
        Test all functionality regarding the 'functions' keyword.
        """
        model = YamlModel()

        function_id = 'f'
        arguments = 'x1, x2'
        formula = 'x1 + x2'

        model.add_function(function_id, arguments, formula)

        # test get_function_by_id
        self.assertIsInstance(model.get_function_by_id(function_id), dict)

        # test overwrite
        with self.assertRaises(ValueError):
            model.add_function(function_id, arguments, formula)

        # now over write function
        model.add_function(function_id, arguments, formula, overwrite=True)

        # test get_function_ids
        self.assertListEqual(model.get_function_ids(),
                             [function_id])

        # test delete
        model.delete_function(function_id)
        self.assertListEqual(model.get_function_ids(), [])

    def test_observable(self):
        """
        Test all functionality regarding the 'observables' keyword.
        """
        model = YamlModel()

        observable_id = 'obs'
        observable_formula = 's*x'
        noise_formula = 'noiseParameter_1'
        observable_name = 'obs_name'

        model.add_observable(observable_id,
                             observable_formula,
                             noise_formula,
                             observable_name=observable_name)

        # test get_observable_by_id
        self.assertIsInstance(model.get_observable_by_id(observable_id), dict)

        # test overwrite
        with self.assertRaises(ValueError):
            model.add_observable(observable_id,
                                 observable_formula,
                                 noise_formula,
                                 observable_name=observable_name)

        # now over write observable
        model.add_observable(observable_id,
                             observable_formula,
                             noise_formula,
                             overwrite=True,
                             observable_name=observable_name)

        # test get_observable_ids
        self.assertListEqual(model.get_observable_ids(),
                             [observable_id])

        # test delete
        model.delete_observable(observable_id)
        self.assertListEqual(model.get_observable_ids(), [])

    def test_condition(self):
        """
        Test all functionality regarding the 'conditions' keyword.
        """
        model = YamlModel()

        condition_id = 'cond'
        condition_name = 'cond'
        condition_dict = {'x1': 42}

        model.add_condition(condition_id,
                            condition_dict,
                            condition_name=condition_name)

        # test get_condition_by_id
        self.assertIsInstance(model.get_condition_by_id(condition_id), dict)

        # test overwrite
        with self.assertRaises(ValueError):
            model.add_condition(condition_id,
                                condition_dict,
                                condition_name=condition_name)

        # now over write condition
        model.add_condition(condition_id,
                            condition_dict,
                            condition_name=condition_name,
                            overwrite=True)

        # test get_condition_ids
        self.assertListEqual(model.get_condition_ids(),
                             [condition_id])

        # test delete
        model.delete_condition(condition_id)
        self.assertListEqual(model.get_condition_ids(), [])

    def test_valid_model(self):
        """Test whether the resulting models are valid."""
        model = YamlModel()

        model.set_time('t')
        model.add_ode(state_id='x',
                      right_hand_side='k_1 * x + t',
                      initial_value=0)
        model.add_parameter(parameter_id='k_1',
                            nominal_value=1)

        model.validate_model()

    def test_minus_in_formula(self):
        """Test writing a model for a right hand side starting with a minus."""
        model = YamlModel()

        model.add_ode(state_id='x',
                      right_hand_side='-x',
                      initial_value=1)

        model.validate_model()

        yaml_dir = os.path.join(self.test_dir, 'test_minus_sign.yaml')
        model.write_to_yaml(yaml_dir)

        validate_yaml(yaml_dir)


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(TestYamlModel())
    unittest.main()
